#include "SSHSession.h"
#include <libssh/callbacks.h>

using SshException = ssh::SshException;

int SSHSession::recv(char *buffer, int size)
{
	try
	{
		return socket.async_read_some(boost::asio::buffer(buffer, size), *yield_);
	}
	catch (const boost::system::system_error &ec)
	{
#ifdef _WIN32
		WSASetLastError(ec.code().value());
		return SOCKET_ERROR;
#else
		errno = ec.code().value();
		return -1;
#endif
	}
}

int SSHSession::send(const char *buffer, int size)
{
	try
	{
		return boost::asio::async_write(socket,
			boost::asio::buffer(buffer, size), *yield_);
	}
	catch (const boost::system::system_error &ec)
	{
#ifdef _WIN32
		WSASetLastError(ec.code().value());
		return SOCKET_ERROR;
#else
		errno = ec.code().value();
		return -1;
#endif
	}
}

void SSHSession::yielder(yield_context &yield)
{
	yield_ = std::make_unique<yield_context>(yield);
}

int SSHSession::closecb()
{
	try
	{
		Disconnect(*yield_);
		return 0;
	}
	catch (const boost::system::system_error &ec)
	{
#ifdef _WIN32
		WSASetLastError(ec.code().value());
		return SOCKET_ERROR;
#else
		errno = ec.code().value();
		return -1;
#endif
	}
}

int SSHSession::srecv(socket_t /*socket*/, void *userdata, char *buffer, int size)
{
	SSHSession *_this = static_cast<SSHSession*>(userdata);
	return _this->recv(buffer, size);
}

int SSHSession::ssend(socket_t /*socket*/, void *userdata, const char *buffer, int size)
{
	SSHSession *_this = static_cast<SSHSession*>(userdata);
	return _this->send(buffer, size);
}

int SSHSession::sclose(socket_t /*socket*/, void *userdata)
{
	SSHSession *_this = static_cast<SSHSession*>(userdata);
	return _this->closecb();
}

SSHSession::SSHSession(io_service &service) : strand_{ service }, TCPSession { service }
{
	setOption(SSH_OPTIONS_FD, socket.native());
	struct ssh_socket_io_callbacks_struct io_callbacks;
	io_callbacks.recv = srecv;
	io_callbacks.send = ssend;
	io_callbacks.closecb = sclose;
	io_callbacks.userdata = this;
	ssh_throw(ssh_set_io_callbacks(c_session, &io_callbacks));
}

std::shared_ptr<SSHSession> SSHSession::Create(io_service &service)
{
	return std::shared_ptr<SSHSession>(new SSHSession(service));
}

void SSHSession::Logon(
	yield_context &yield,
	const char *host, unsigned short port,
	const char *username, const char *passwd)
{
	Connect(yield, host, port);
	setOption(SSH_OPTIONS_FD, socket.native());
	setOption(SSH_OPTIONS_USER, username);
	yielder(yield);
	connect();
	if (userauthPublickeyAuto() != SSH_AUTH_SUCCESS &&
		userauthPassword(passwd) != SSH_AUTH_SUCCESS)
		throw std::system_error{ ERROR_INVALID_SERVICE_ACCOUNT, std::system_category() };
}

sftp_channel_ptr SSHSession::CreateSFTP(yield_context &yield)
{
	yielder(yield);
	sftp_session sftp = sftp_new(c_session);
	throw_null(sftp);
	return sftp_channel_ptr{ new SFTPChannel(std::static_pointer_cast<SSHSession>(shared_from_this()), sftp) };
}


std::future<void> SSHSession::Logon(
	const char *host, unsigned short port,
	const char *username, const char *passwd)
{
	return when(strand_, this, &SSHSession::Logon, host, port, username, passwd);
}

std::future<sftp_channel_ptr> SSHSession::CreateSFTP()
{
	return when(strand_, this, &SSHSession::CreateSFTP);
}

ssh_session SSHSession::GetSession()
{
	return c_session;
}

void SSHSession::throw_null(void *ptr)
{
	if (!ptr)
	{
		throw SshException(GetSession());
	}
}

bool SSHSession::throw_error(ssize_t error)
{
	if (error == SSH_ERROR)
	{
		throw SshException(GetSession());
	}
	else
		return true;
}

SFTPFile::SFTPFile(sftp_channel_ptr channel, sftp_file file)
	:channel_(channel), file_(file)
{
}

SFTPFile::~SFTPFile()
{
	sftp_close(file_);
}

SFTPDirectory::SFTPDirectory(sftp_channel_ptr channel, sftp_dir dir)
	:channel_(channel), dir_(dir)
{

}

SFTPDirectory::~SFTPDirectory()
{
	sftp_closedir(dir_);
}

SFTPChannel::SFTPChannel(std::shared_ptr<SSHSession> session, sftp_session sftp)
	: session_(session), sftp_(sftp), last_status_(0)
{

}

SFTPChannel::~SFTPChannel()
{
	sftp_free(sftp_);
}

void SFTPChannel::throw_exception()
{
	if (ssh_get_error_code(session_->GetSession()) == SSH_REQUEST_DENIED)
	{
		last_status_ = sftp_get_error(sftp_);
		last_message_ = ssh_get_error(session_->GetSession());
	}
	else throw SshException(session_->GetSession());
}

void SFTPChannel::throw_null(void *ptr)
{
	if (!ptr)
	{
		throw_exception();
	}
}

bool SFTPChannel::throw_error(ssize_t error)
{
	if (error == SSH_ERROR)
	{
		throw_exception();
		return false;
	}
	else
		return true;
}

sftp_file_ptr SFTPChannel::OpenFile(yield_context &_yield, const char *filename, int access, mode_t mode)
{
	session_->yielder(_yield);
	sftp_file file = sftp_open(sftp_, filename, access, mode);
	throw_null(file);
	return std::make_unique<SFTPFile>(shared_from_this(), file);
}

sftp_dir_ptr SFTPChannel::OpenDir(yield_context &_yield, const char *path)
{
	session_->yielder(_yield);
	sftp_dir dir = sftp_opendir(sftp_, path);
	throw_null(dir);
	return std::make_unique<SFTPDirectory>(shared_from_this(), dir);
}


std::future<sftp_file_ptr> SFTPChannel::OpenFile(const char *filename, int access, mode_t mode)
{
	return when(session_->strand_, this, &SFTPChannel::OpenFile, filename, access, mode);
}

std::future<sftp_dir_ptr> SFTPChannel::OpenDir(const char *path)
{
	return when(session_->strand_, this, &SFTPChannel::OpenDir, path);
}
